
import numpy as np
import scipy.sparse.linalg as sp

MAX_IT = 100 # Maximum number of ipm iterations
MAX_IT_cg = 1000 # Maximum number of cg iterations


def error_adjusted_ipm(A_mat, initial_point, ipm_tol, solveV_tol, sketch_size = 0):
    '''
    Error-adjusted predictor corrector algorithm (Algorithm 1) in the paper.

    Parameters
    ----------
    A_mat : np.array[,]
        LP constraint matrix.
    initial_point : tuple
        The intial point is a primal-dual pair (x_init, y_init, s_init) where
        each item in the tuple is a numpy.array of the correct length.
    ipm_tol : float
        Error tolerance of the IPM (epsilon).  
        Will return once the duality measure is less than ipm_tol.
    solveV_tol : float
        Error of SolveV (delta).
    sketch_size : int, optional
        If sketch_size equals zero, then solveV_perturb will be used for the 
        linear solver.  Otherwise, the PCG-based solver will be used with the given
        sketch size (w). The default is 0.

    Returns
    -------
    results_dict : dict
        This function returns a dictionary containing: The duality measure at 
        each iteration, the number of PCG iterations in the corrector step at
        each outer iteration, the final number of outer iterations, and the
        final primal infeasibility.

    '''
    
    
    # Extract objectives/constraints from implicit LP
    x_vec, y_vec, s_vec = initial_point
    b_vec = A_mat @ x_vec
    c_vec = A_mat.T @ y_vec + s_vec
    
    n = x_vec.shape[0] #    
    iteration_count = 0
    
    results_dict = dict()
    results_dict['mu_list'] = [(1/n) * x_vec.T @ s_vec]
    results_dict['inner_it'] = []
    
    while (1/n) * x_vec.T @ s_vec > ipm_tol:
        
        iteration_count += 1
        if iteration_count > MAX_IT:
            break
        
        mu = (1/n) * x_vec.T @ s_vec
        
        D_square_mat = np.diag(x_vec/s_vec)
        
        if sketch_size == 0:
            p_vec = A_mat @ x_vec
            y_step, v_vec = solveV_perturb(A_mat, D_square_mat, p_vec, x_vec, 
                                           s_vec, solveV_tol) 
        else:
            p_vec = A_mat @ x_vec
            y_step, v_vec, pcg_pred_it = solveV_pcg(A_mat, D_square_mat, p_vec, x_vec, 
                                       s_vec, solveV_tol, sketch_size) 
                
        
        
        s_step = -1 * A_mat.T @ y_step
        x_step = -x_vec - D_square_mat @ s_step - (1/s_vec) * v_vec
        
        alpha = min(0.5, (mu/(16*np.linalg.norm(x_step * s_step)))**(1/2))
        
        x_vec += alpha * x_step
        y_vec += alpha * y_step
        s_vec += alpha * s_step
            
        
        # Corrector Step
        
        mu = (1/n) * x_vec.T @ s_vec
        D_square_mat = np.diag(x_vec/s_vec)
        
        if sketch_size == 0:
        
            p_vec = -1 * mu * A_mat @ (1/s_vec) + A_mat @ x_vec
            y_step, v_vec = solveV_perturb(A_mat, D_square_mat, p_vec, x_vec, 
                                           s_vec, solveV_tol) 
            
        else:
            p_vec = -1 * mu * A_mat @ (1/s_vec) + A_mat @ x_vec
            y_step, v_vec, pcg_corr_it = solveV_pcg(A_mat, D_square_mat, p_vec, x_vec, 
                                       s_vec, solveV_tol, sketch_size) 
            results_dict['inner_it'].append([pcg_pred_it, pcg_corr_it])
            
            
        s_step = -1 * A_mat.T @ y_step
        x_step = -x_vec + mu * (1/s_vec) - D_square_mat @ s_step - (1/s_vec) * v_vec

        x_vec += x_step
        y_vec += y_step
        s_vec += s_step
        
        results_dict['mu_list'].append((1/n) * x_vec.T @ s_vec)
        
    results_dict['num_iterations'] = iteration_count
    results_dict['primal_infeasibilty'] = np.linalg.norm(A_mat @ x_vec - b_vec)
    return results_dict
        
        
def solveV_perturb(A_mat, D_square_mat, p_vec, x_vec, s_vec, solveV_tol):
    ##  Maximally perturb linear system solve to fulifll SolveV guarantee 
    ##  with accuracy parameter delta = solveV_tol

    coeff_mat = A_mat @ D_square_mat @ A_mat.T
    
    v_vec = np.random.normal(size=s_vec.shape[0])
    v_vec = (1/np.linalg.norm(v_vec)) * v_vec
    v_vec = solveV_tol * v_vec
    
    
    pert_sol = np.linalg.lstsq(coeff_mat,
                               p_vec + A_mat @ (1/s_vec * v_vec),
                               rcond = None)[0]

    return pert_sol, v_vec
    
def solveV_pcg(A_mat, D_square_mat, p_vec, x_vec, s_vec, solveV_tol, sketch_size):
    ##  Use PCG (Algorithm 3) linear system solve to fulifll SolveV guarantee 
    ##  with accuracy parameter delta = solveV_tol
    
    W_mat = np.random.normal(0, pow(sketch_size, -1/2),
                                     size = (D_square_mat.shape[0], sketch_size))
    
    U_mat, Sig_mat, _ =  np.linalg.svd(A_mat @ (D_square_mat**(1/2)) @ W_mat)
    
    Q_mat_inv = U_mat @ np.diag(Sig_mat**(-1)) @ U_mat.T
    
    coeff_mat = A_mat @ D_square_mat @ A_mat.T
    
    
    # Defining callback function to count number of pcg iterations    
    pcg_iters = 0
    def callback(xk):
         nonlocal pcg_iters
         pcg_iters+=1

    
    y_step, _ = sp.cg(coeff_mat, p_vec, tol=solveV_tol, maxiter=MAX_IT_cg,
                      M = Q_mat_inv, callback = callback)
    
    
    residual_vec = A_mat @ D_square_mat @ A_mat.T @ y_step - p_vec
    
    v_vec = np.diag(x_vec * s_vec)**(1/2) @ W_mat \
        @ np.linalg.pinv(A_mat @ D_square_mat**(1/2) @ W_mat) @ residual_vec

    return y_step, v_vec, pcg_iters





