
import numpy as np

def construct_random_LP(m_dim, n_dim, mu_init = 20, entry_bound = 10):
    '''
    This function finds an initial primal-dual point in some LP with 'n_dim' variables
    and 'm_dim' constraints.  The initial point is primal-dual feasible and
    and on the central path of the implicit LP,  The constraint vector 'b_vec' and
    objective vector 'c_vec' and be computed from the returned values.

    Parameters
    ----------
    m_dim : int
        Number of constraints in the linear program.
    n_dim : int
        Number of variables in the linear program.
    mu_init : float, optional
        The duality measure of the returned point. The default is 20.
    entry_bound : float, optional
        Maximum magnitude of entries in randomly generated coefficient matrix and
        initial points. The default is 10.
        
    Returns
    -------
    initial_point : tuple
        The intial point is a primal-dual pair (x_init, y_init, s_init) where
        each item in the tuple is a numpy.array of the correct length.
    A_mat : np.array[m_dim, n_dim]
        Randomly generated constraint matrix.

    '''
    
    # sampled starting at 0.01 cuz numerical errors are giving negative values
        
    A_mat = np.random.uniform(-1*entry_bound, entry_bound, size = (m_dim, n_dim))
    x_init = np.random.uniform(0.01, entry_bound, size = n_dim)
    y_init = np.random.uniform(-1*entry_bound, entry_bound, size = m_dim)
    s_init = (1/x_init)
    x_init = mu_init*x_init

    initial_point = (x_init, y_init, s_init)
    return A_mat, initial_point