
import math

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from error_adj_ipm import error_adjusted_ipm
from generate_lp import construct_random_LP


'''
This code tests the relationship between the number of iterations of the error-adjusted
IPM and the desired duality measure (epsilon) of the returned solution. Run to 
reproduce Figure 2 or 4 depending on SKETCH_SIZE.
'''

FILENAME = 'num_it_vs_eps.png'  # Output file of plot
NUM_CONSTRAINTS = 30            # Number of rows in the constraint matrix 'A'
NUM_VAR = 70                    # Number of columns in the constraint matrix 'A'
REPS = 60                       # Number of times to repeat the experiment at each 'n'
SKETCH_SIZE = 0                # Size of sketch used in iterative solver.  Set to zero for SolveV-perturb
eps_list = [0.1, 0.01, 0.001, 1E-4, 1E-5, 1E-6, 1E-7, 1E-8]  # List of values of epsilon to test.
delta_func = lambda eps: eps    # Precision of linear solver as a function of 'n'


''' Test convergence dependency on epsilon '''

result_list = []

for eps in eps_list:
    print("Current eps:", eps)
    for rep_it in range(REPS):
        print("Rep:", rep_it + 1)
        
        A_mat, initial_point = construct_random_LP(NUM_CONSTRAINTS, NUM_VAR)
        results = error_adjusted_ipm(A_mat, initial_point, eps,
                                     delta_func(eps), sketch_size = SKETCH_SIZE)
        
        result_list.append([eps,
                            results['num_iterations'],
                            results['primal_infeasibilty']])
        
result_df = pd.DataFrame(result_list)
result_df.columns = ['eps', 'num_it', 'primal_infeas']

result_df = result_df.groupby('eps').quantile([0.1, 0.5, 0.9]).unstack()['num_it'].reset_index()
trans_df = result_df.copy()

trans_df['eps'] = trans_df['eps'].apply(lambda x: -1*math.log(x, 10))

plt.xlabel("log(1/eps)")
plt.ylabel("Number iterations")


plt.plot(trans_df['eps'], trans_df[0.5], 'ro-', markersize = 5)
plt.fill_between(trans_df['eps'], trans_df[0.1], trans_df[0.9], color='blue', alpha=0.5)

plt.savefig(FILENAME, dpi=300)
